@@ -5,8 +5,8 @@
 ;; Author: Wilfred Hughes <me@wilfred.me.uk>
 ;; URL: https://github.com/Wilfred/helpful
 ;; Keywords: help, lisp
-;; Version: 0.16
-;; Package-Requires: ((emacs "25.1") (dash "2.12.0") (dash-functional "1.2.0") (s "1.11.0") (f "0.20.0") (elisp-refs "1.2") (shut-up "0.3"))
+;; Version: 0.17
+;; Package-Requires: ((emacs "25") (dash "2.12.0") (dash-functional "1.2.0") (s "1.11.0") (f "0.20.0") (elisp-refs "1.2"))
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -49,7 +49,6 @@
 (require 'dash-functional)
 (require 's)
 (require 'f)
-(require 'shut-up)
 (require 'find-func)
 (require 'nadvice)
 (require 'info-look)
@@ -67,6 +66,12 @@ buffer-local variables.")
 (defvar-local helpful--view-literal nil
   "Whether to show a value as a literal, or a pretty interactive
 view.")
+(defvar-local helpful--first-display t
+  "Whether this is the first time this results buffer has been
+displayed.
+
+Nil means that we're refreshing, so we don't want to clobber any
+settings changed by the user.")
 
 (defgroup helpful nil
   "A rich help system with contextual information."
@@ -757,51 +762,34 @@ blank line afterwards."
                 (-cons* first-line "" (cdr lines)))
       docstring)))
 
-(defun helpful--propertize-keywords (docstring)
-  "Propertize quoted keywords in docstrings."
-  (replace-regexp-in-string
-   ;; Replace all text of the form `foo'.
-   (rx "`"
-       (group ":"
-              symbol-start
-              (+? (or (syntax word) (syntax symbol)))
-              symbol-end)
-       "'")
-   (lambda (it)
-     (propertize (match-string 1 it)
-                 'face 'font-lock-builtin-face))
-   docstring
-   t t))
-
-(defun helpful--propertize-quoted (docstring)
-  "Convert `foo' in docstrings to buttons (if bound) or else highlight."
-  (replace-regexp-in-string
-   ;; Replace all text of the form `foo'.
-   (rx (? "\\=") "`" (+? (not (any "`" "'"))) "'")
-   (lambda (it)
-     (let* ((sym-name
-             (s-chop-prefix "`" (s-chop-suffix "'" it)))
-            (sym (intern sym-name)))
-       (cond
-        ;; If the quote is escaped, don't modify it.
-        ((s-starts-with-p "\\=" it)
-         it)
-        ;; Only create a link if this is a symbol that is bound as a
-        ;; variable or callable.
-        ((or (boundp sym) (fboundp sym))
-         (helpful--button
-          sym-name
-          'helpful-describe-button
-          'symbol sym))
-        ;; If this is already a button, don't modify it.
-        ((get-text-property 0 'button sym-name)
-         sym-name)
-        ;; Highlight the quoted string.
-        (t
-         (propertize sym-name
-                     'face 'font-lock-constant-face)))))
-   docstring
-   t t))
+(defun helpful--propertize-sym-ref (sym-name)
+  "Given a symbol name from a docstring, convert to a button (if
+bound) or else highlight."
+  (let* ((sym (intern sym-name)))
+    (cond
+     ;; Highlight keywords.
+     ((s-matches-p
+       (rx ":"
+           symbol-start
+           (+? (or (syntax word) (syntax symbol)))
+           symbol-end)
+       sym-name)
+      (propertize sym-name
+                  'face 'font-lock-builtin-face))
+     ;; Only create a link if this is a symbol that is bound as a
+     ;; variable or callable.
+     ((or (boundp sym) (fboundp sym))
+      (helpful--button
+       sym-name
+       'helpful-describe-button
+       'symbol sym))
+     ;; If this is already a button, don't modify it.
+     ((get-text-property 0 'button sym-name)
+      sym-name)
+     ;; Highlight the quoted string.
+     (t
+      (propertize sym-name
+                  'face 'font-lock-constant-face)))))
 
 (defun helpful--propertize-info (docstring)
   "Convert info references in docstrings to buttons."
@@ -857,7 +845,10 @@ vector suitable for `key-description', and COMMAND is a smbol."
    ;; so this is a command we can call.
    ((or
      (symbolp keymap)
-     (functionp keymap))
+     (functionp keymap)
+     ;; Strings or vectors mean a keyboard macro.
+     (stringp keymap)
+     (vectorp keymap))
     `(([] ,keymap)))
    ((stringp (car keymap))
     (helpful--keymap-keys (cdr keymap)))
@@ -934,12 +925,16 @@ vector suitable for `key-description', and COMMAND is a smbol."
                          keys-and-commands))
          (formatted-commands
           (--map
-           (if (symbolp it)
-               (helpful--button
-                (symbol-name it)
-                'helpful-describe-button
-                'symbol it)
-             "#<anonymous-function>")
+           (cond
+            ((symbolp it)
+             (helpful--button
+              (symbol-name it)
+              'helpful-describe-button
+              'symbol it))
+            ((or (stringp it) (vectorp it))
+             "Keyboard Macro")
+            (t
+             "#<anonymous-function>"))
            commands))
          ;; Build lines for display.
          (lines
@@ -949,13 +944,35 @@ vector suitable for `key-description', and COMMAND is a smbol."
     ;; inherited bindings last. Sort so that we group by prefix.
     (s-join "\n" (-sort #'string< lines))))
 
+(defun helpful--format-commands (str keymap)
+  "Replace all the \\[ references in STR with buttons."
+  (replace-regexp-in-string
+   ;; Text of the form \\[foo-command]
+   (rx "\\[" (group (+ (not (in "]")))) "]")
+   (lambda (it)
+     (let* ((symbol-name (match-string 1 it))
+            (symbol (intern symbol-name))
+            (key (where-is-internal symbol keymap t))
+            (key-description
+             (if key
+                 (key-description key)
+               (format "M-x %s" symbol-name))))
+       (helpful--button
+        key-description
+        'helpful-describe-exactly-button
+        'symbol symbol
+        'callable-p t)))
+   str
+   t
+   t))
+
 (defun helpful--format-command-keys (docstring)
   "Convert command key references and keymap references
 in DOCSTRING to buttons.
 
 Emacs uses \\= to escape \\[ references, so replace that
 unescaping too."
-  ;; Based on `substitute-command-keys', but converts command
+  ;; Loosely based on `substitute-command-keys', but converts
   ;; references to buttons.
   (let ((keymap nil))
     (with-temp-buffer
@@ -963,6 +980,20 @@ unescaping too."
       (goto-char (point-min))
       (while (not (eobp))
         (cond
+         ((looking-at
+           ;; Text of the form "foo"
+           (rx "\""))
+          ;; For literal strings, escape backslashes so our output
+          ;; shows copy-pasteable literals.
+          (let* ((start-pos (point))
+                 (end-pos (progn (forward-char) (search-forward "\"" nil t)))
+                 contents)
+            (if end-pos
+                (progn
+                  (setq contents (buffer-substring start-pos end-pos))
+                  (delete-region start-pos end-pos)
+                  (insert (s-replace "\\" "\\\\" contents)))
+              (forward-char 1))))
          ((looking-at
            ;; Text of the form \=X
            (rx "\\="))
@@ -970,9 +1001,33 @@ unescaping too."
           ;; Step over the escaped character.
           (delete-region (point) (+ (point) 2))
           (forward-char 1))
+         ((looking-at
+           ;; Text of the form `foo'
+           (rx "`"))
+          (let* ((start-pos (point))
+                 (end-pos (search-forward "'" nil t))
+                 (contents
+                  (when end-pos
+                    (buffer-substring (1+ start-pos) (1- end-pos)))))
+            (cond
+             ((null contents)
+              ;; If there's no closing ' to match the opening `, just
+              ;; leave it.
+              (goto-char (1+ start-pos)))
+             ((s-contains-p "`" contents)
+              ;; If we have repeated backticks `foo `bar', leave the
+              ;; first one.
+              (goto-char (1+ start-pos)))
+             ((s-contains-p "\\[" contents)
+              (delete-region start-pos end-pos)
+              (insert (helpful--format-commands contents keymap)))
+             (t
+              (delete-region start-pos end-pos)
+              (insert (helpful--propertize-sym-ref contents))))))
          ((looking-at
            ;; Text of the form \\<foo-keymap>
-           (rx "\\<" (group (+ (not (in ">")))) ">"))
+           (rx "\\<" (group (+ (not (in ">")))) ">"
+               (? "\n")))
           (let* ((symbol-with-parens (match-string 0))
                  (symbol-name (match-string 1)))
             ;; Remove the original string.
@@ -999,24 +1054,12 @@ unescaping too."
          ((looking-at
            ;; Text of the form \\[foo-command]
            (rx "\\[" (group (+ (not (in "]")))) "]"))
-          (let* ((symbol-with-parens (match-string 0))
-                 (symbol-name (match-string 1)))
+          (let* ((symbol-with-parens (match-string 0)))
             ;; Remove the original string.
             (delete-region (point)
                            (+ (point) (length symbol-with-parens)))
             ;; Add a button.
-            (let* ((symbol (intern symbol-name))
-                   (key (where-is-internal symbol keymap t))
-                   (key-description
-                    (if key
-                        (key-description key)
-                      (format "M-x %s" symbol-name))))
-              (insert
-               (helpful--button
-                key-description
-                'helpful-describe-exactly-button
-                'symbol symbol
-                'callable-p t)))))
+            (insert (helpful--format-commands symbol-with-parens keymap))))
          ;; Don't modify other characters.
          (t
           (forward-char 1))))
@@ -1031,10 +1074,6 @@ unescaping too."
       (helpful--propertize-info)
       (helpful--propertize-links)
       (helpful--propertize-bare-links)
-      (helpful--propertize-keywords)
-      (helpful--propertize-quoted)
-      ;; This needs to happen after we've replaced quoted chars, so we
-      ;; don't confuse \\=` with `.
       (helpful--format-command-keys)
       (s-trim)))
 
@@ -1187,7 +1226,8 @@ If the source code cannot be found, return the sexp used."
 (defun helpful--in-manual-p (sym)
   "Return non-nil if SYM is in an Info manual."
   (let ((completions
-         (shut-up
+         (cl-letf (((symbol-function #'message)
+                    (lambda (_format-string &rest _args))))
            (info-lookup->completions 'symbol 'emacs-lisp-mode))))
     (-when-let (buf (get-buffer " temp-info-look"))
       (kill-buffer buf))
@@ -1277,8 +1317,9 @@ buffer."
         (buf nil)
         (pos nil)
         (opened nil)
-        ;; Skip running find-file-hook since it may prompt the user.
+        ;; Skip running hooks that may prompt the user.
         (find-file-hook nil)
+        (after-change-major-mode-hook nil)
         ;; If we end up opening a buffer, don't bother with file
         ;; variables. It prompts the user, and we discard the buffer
         ;; afterwards anyway.
@@ -1725,6 +1766,11 @@ OBJ may be a symbol or a compiled function object."
     (setq file-name (s-chop-suffix ".gz" file-name))
     (help-fns--autoloaded-p sym file-name)))
 
+(defun helpful--compiled-p (sym)
+  "Return non-nil if function SYM is byte-compiled"
+  (and (symbolp sym)
+       (byte-code-function-p (symbol-function sym))))
+
 (defun helpful--summary (sym callable-p buf pos)
   "Return a one sentence summary for SYM."
   (-let* ((primitive-p (helpful--primitive-p sym callable-p))
@@ -1763,6 +1809,11 @@ OBJ may be a symbol or a compiled function object."
             "autoloaded"
             'helpful-info-button
             'info-node "(elisp)Autoload"))
+          (compiled-button
+           (helpful--button
+            "compiled"
+            'helpful-info-button
+            'info-node "(elisp)Byte Compilation"))
           (buffer-local-button
            (helpful--button
             "buffer-local"
@@ -1770,24 +1821,32 @@ OBJ may be a symbol or a compiled function object."
             'info-node "(elisp)Buffer-Local Variables"))
           (autoloaded-p
            (and callable-p buf (helpful--autoloaded-p sym buf)))
+          (compiled-p
+           (and callable-p (helpful--compiled-p sym)))
           (description
-           (cond
-            (alias-p
-             (format "%s %s"
-                     (if callable-p "a" "an")
-                     alias-button))
-            ((and callable-p (commandp sym) autoloaded-p)
-             (format "an %s, %s" interactive-button autoload-button))
-            ((helpful--kbd-macro-p sym) "a")
-            ((and callable-p (commandp sym))
-             (format "an %s" interactive-button))
-            ((and callable-p autoloaded-p)
-             (format "an %s" autoload-button))
-            ((and (not callable-p)
-                  (local-variable-if-set-p sym))
-             (format "a %s" buffer-local-button))
-            (t
-             "a")))
+           (concat
+            (cond
+             (alias-p
+              (format "%s %s"
+                      (if callable-p "a" "an")
+                      alias-button))
+             ((and callable-p (commandp sym) autoloaded-p)
+              (format "an %s, %s" interactive-button autoload-button))
+             ((helpful--kbd-macro-p sym) "a")
+             ((and callable-p (commandp sym))
+              (format "an %s" interactive-button))
+             ((and callable-p autoloaded-p)
+              (format "an %s" autoload-button))
+             ((and (not callable-p)
+                   (local-variable-if-set-p sym))
+              (format "a %s" buffer-local-button))
+             (t
+              "a"))
+            (if compiled-p
+                (format "%s %s"
+                        (if (or (commandp sym) autoloaded-p) "," "")
+                        compiled-button))
+            ""))
           (kind
            (cond
             ((special-form-p sym)
@@ -1934,6 +1993,34 @@ may contain duplicates."
       (when (autoloadp fn-obj)
         (autoload-do-load fn-obj)))))
 
+(defun helpful--hook-p (symbol value)
+  "Does SYMBOL look like a hook?"
+  (and
+   (or
+    (s-ends-with-p "-hook" (symbol-name symbol))
+    ;; E.g. `after-change-functions', which can be used with
+    ;; `add-hook'.
+    (s-ends-with-p "-functions" (symbol-name symbol)))
+   (consp value)))
+
+(defun helpful--format-value (sym value)
+  "Format VALUE as a string."
+  (cond
+   (helpful--view-literal
+    (helpful--syntax-highlight (helpful--pretty-print value)))
+   ;; Allow strings to be viewed with properties rendered in
+   ;; Emacs, rather than as a literal.
+   ((stringp value)
+    value)
+   ;; Allow keymaps to be viewed with keybindings shown and
+   ;; links to the commands bound.
+   ((keymapp value)
+    (helpful--format-keymap value))
+   ((helpful--hook-p sym value)
+    (helpful--format-hook value))
+   (t
+    (helpful--pretty-print value))))
+
 (defun helpful-update ()
   "Update the current *Helpful* buffer to the latest
 state of the current symbol."
@@ -1989,8 +2076,14 @@ state of the current symbol."
              (multiple-views-p
               (or (stringp val)
                   (keymapp val)
-                  (and (s-ends-with-p "-hook" (symbol-name sym))
-                       (consp val)))))
+                  (helpful--hook-p sym val))))
+        (when helpful--first-display
+          (if (stringp val)
+              ;; For strings, it's more intuitive to display them as
+              ;; literals, so "1" and 1 are distinct.
+              (setq helpful--view-literal t)
+            ;; For everything else, prefer the pretty view if available.
+            (setq helpful--view-literal nil)))
         (insert
          (helpful--heading
           (cond
@@ -2010,22 +2103,7 @@ state of the current symbol."
             "Global Value")
            ;; This variable is not buffer-local.
            (t "Value")))
-         (cond
-          (helpful--view-literal
-           (helpful--syntax-highlight (helpful--pretty-print val)))
-          ;; Allow strings to be viewed with properties rendered in
-          ;; Emacs, rather than as a literal.
-          ((stringp val)
-           val)
-          ;; Allow keymaps to be viewed with keybindings shown and
-          ;; links to the commands bound.
-          ((keymapp val)
-           (helpful--format-keymap val))
-          ((and (s-ends-with-p "-hook" (symbol-name sym))
-                (consp val))
-           (helpful--format-hook val))
-          (t
-           (helpful--pretty-print val)))
+         (helpful--format-value sym val)
          "\n\n")
         (when multiple-views-p
           (insert (helpful--make-toggle-literal-button) " "))
@@ -2158,32 +2236,31 @@ state of the current symbol."
 
     (helpful--insert-section-break)
 
-    (insert
-     (helpful--heading
-      (if (eq helpful--sym canonical-sym)
-          "Source Code"
-        "Alias Source Code"))
-     (cond
-      (source-path
-       (concat
-        (propertize (format "%s Defined in " (if primitive-p "//" ";;"))
-                    'face 'font-lock-comment-face)
-        (helpful--navigate-button
-         (f-abbrev source-path)
-         source-path
-         pos)
-        "\n"))
-      (primitive-p
-       (concat
-        (propertize
-         "C code is not yet loaded."
-         'face 'font-lock-comment-face)
-        "\n\n"
-        (helpful--button
-         "Set C source directory"
-         'helpful-c-source-directory)))
-      (t
-       "")))
+    (when (or source-path primitive-p)
+      (insert
+       (helpful--heading
+        (if (eq helpful--sym canonical-sym)
+            "Source Code"
+          "Alias Source Code"))
+       (cond
+        (source-path
+         (concat
+          (propertize (format "%s Defined in " (if primitive-p "//" ";;"))
+                      'face 'font-lock-comment-face)
+          (helpful--navigate-button
+           (f-abbrev source-path)
+           source-path
+           pos)
+          "\n"))
+        (primitive-p
+         (concat
+          (propertize
+           "C code is not yet loaded."
+           'face 'font-lock-comment-face)
+          "\n\n"
+          (helpful--button
+           "Set C source directory"
+           'helpful-c-source-directory))))))
     (when source
       (insert
        (cond
@@ -2216,6 +2293,7 @@ state of the current symbol."
     (goto-char (point-min))
     (forward-line (1- start-line))
     (forward-char start-column)
+    (setq helpful--first-display nil)
 
     (when opened
       (kill-buffer buf))))
@@ -2288,7 +2366,7 @@ For example, \"(some-func FOO &optional BAR)\"."
 
 (defun helpful--format-obsolete-info (sym callable-p)
   (-let [(use _ date) (helpful--obsolete-info sym callable-p)]
-    (helpful--propertize-quoted
+    (helpful--format-docstring
      (s-word-wrap
       70
       (format "This %s is obsolete%s%s"
@@ -2404,6 +2482,25 @@ or :foo."
   (or (fboundp symbol)
       (helpful--variable-p symbol)))
 
+(defun helpful--bookmark-jump (bookmark)
+  "Create and switch to helpful bookmark BOOKMARK."
+  (let ((callable-p (bookmark-prop-get bookmark 'callable-p))
+        (sym (bookmark-prop-get bookmark 'sym))
+        (position (bookmark-prop-get bookmark 'position)))
+    (if callable-p
+        (helpful-callable sym)
+      (helpful-variable sym))
+    (goto-char position)))
+
+(defun helpful--bookmark-make-record ()
+  "Create a bookmark record for helpful buffers.
+
+See docs of `bookmark-make-record-function'."
+  `((sym . ,helpful--sym)
+    (callable-p . ,helpful--callable-p)
+    (position    . ,(point))
+    (handler     . helpful--bookmark-jump)))
+
 (defun helpful--convert-c-name (symbol var)
   "Convert SYMBOL from a C name to an Elisp name.
 E.g. convert `Fmake_string' to `make-string' or
@@ -2425,27 +2522,25 @@ nil if SYMBOL doesn't begin with \"F\" or \"V\"."
 See also `helpful-callable' and `helpful-variable'."
   (interactive
    (list (helpful--read-symbol "Symbol: " #'helpful--bound-p)))
-  (cond
-   ((and (boundp symbol) (fboundp symbol))
-    (if (y-or-n-p
-         (format "%s is a both a variable and a callable, show variable?"
-                 symbol))
-        (helpful-variable symbol)
-      (helpful-callable symbol)))
-   ((fboundp symbol)
-    (helpful-callable symbol))
-   ((boundp symbol)
-    (helpful-variable symbol))
-   ((let ((fsym (helpful--convert-c-name symbol nil)))
-      (when (fboundp fsym)
-        (helpful-callable fsym)
-        fsym)))
-   ((let ((vsym (helpful--convert-c-name symbol t)))
-      (when (boundp vsym)
-        (helpful-variable vsym)
-        vsym)))
-   (t
-    (user-error "Not bound: %S" symbol))))
+  (let ((c-var-sym (helpful--convert-c-name symbol nil))
+        (c-fn-sym (helpful--convert-c-name symbol t)))
+    (cond
+     ((and (boundp symbol) (fboundp symbol))
+      (if (y-or-n-p
+           (format "%s is a both a variable and a callable, show variable?"
+                   symbol))
+          (helpful-variable symbol)
+        (helpful-callable symbol)))
+     ((fboundp symbol)
+      (helpful-callable symbol))
+     ((boundp symbol)
+      (helpful-variable symbol))
+     ((and c-fn-sym (fboundp c-fn-sym))
+      (helpful-callable c-fn-sym))
+     ((and c-var-sym (boundp c-var-sym))
+      (helpful-variable c-var-sym))
+     (t
+      (user-error "Not bound: %S" symbol)))))
 
 ;;;###autoload
 (defun helpful-variable (symbol)
@@ -2544,13 +2639,25 @@ See also `helpful-max-buffers'."
     map)
   "Keymap for `helpful-mode'.")
 
+(declare-function bookmark-prop-get "bookmark" (bookmark prop))
+(declare-function bookmark-make-record-default "bookmark"
+                  (&optional no-file no-context posn))
+;; Ensure this variable is defined even if bookmark.el isn't loaded
+;; yet. This follows the pattern in help-mode.el.gz.
+;; TODO: find a cleaner solution.
+(defvar bookmark-make-record-function)
+
 (define-derived-mode helpful-mode special-mode "Helpful"
   "Major mode for *Helpful* buffers."
   (add-hook 'xref-backend-functions #'elisp--xref-backend nil t)
 
   (setq imenu-create-index-function #'helpful--imenu-index)
   ;; Prevent imenu converting "Source Code" to "Source.Code".
-  (setq-local imenu-space-replacement " "))
+  (setq-local imenu-space-replacement " ")
+
+  ;; Enable users to bookmark helpful buffers.
+  (set (make-local-variable 'bookmark-make-record-function)
+       #'helpful--bookmark-make-record))
 
 (provide 'helpful)
 ;;; helpful.el ends here
