@@ -56,7 +56,11 @@
   ;; We should handle stray backquotes.
   (let* ((formatted-docstring (helpful--format-docstring "`foo `message'")))
     (should
-     (equal formatted-docstring "`foo message"))))
+     (equal formatted-docstring "`foo message")))
+  ;; Handle a missing closing '.
+  (let* ((formatted-docstring (helpful--format-docstring "`foo")))
+    (should
+     (equal formatted-docstring "`foo"))))
 
 (ert-deftest helpful--docstring-unescape ()
   "Discard \\=\\= in docstrings."
@@ -65,6 +69,28 @@
     (should
      (not (s-contains-p "\\=" formatted-docstring)))))
 
+(ert-deftest helpful--docstring-strings ()
+  "Double-quoted strings should be treated literally."
+  ;; Ensure backslashes are shown escaped, so the output is a valid string literal.
+  (let* ((formatted-docstring
+          (helpful--format-docstring
+           "hello \"x\\y\" world")))
+    ;; This test will fail in a local Emacs instance that has modified
+    ;; minibuffer keybindings.
+    (should
+     (string-equal formatted-docstring "hello \"x\\\\y\" world")))
+  ;; Don't crash on unbalanced doublequotes.
+  (helpful--format-docstring "hello \" world")
+  ;; Command sequences \\[foo] should be ignored inside doublequotes.
+  (let* ((formatted-docstring
+          (helpful--format-docstring
+           "hello \"\\[foo]\" world")))
+    ;; This test will fail in a local Emacs instance that has modified
+    ;; minibuffer keybindings.
+    (should
+     (string-equal formatted-docstring
+                   "hello \"\\\\[foo]\" world"))))
+
 (ert-deftest helpful--docstring-keymap ()
   "Handle keymap references in docstrings."
   (let* ((formatted-docstring
@@ -75,6 +101,17 @@
     (should
      (string-equal formatted-docstring "M-n"))))
 
+(ert-deftest helpful--docstring-keymap-newline ()
+  "If a keymap reference is on its own line, remove the entire line."
+  (should
+   (string-equal
+    (helpful--format-docstring "Foo.
+\\<minibuffer-local-map>
+bar")
+    "Foo.
+
+bar")))
+
 (ert-deftest helpful--docstring-advice ()
   "Get the docstring on advised functions."
   (should
@@ -103,7 +140,8 @@
     (with-temp-buffer
       (insert "(defun test-foo-edebug () 44)")
       (goto-char (point-min))
-      (shut-up
+      (cl-letf (((symbol-function #'message)
+                 (lambda (_format-string &rest _args))))
         (eval (eval-sexp-add-defvars (edebug-read-top-level-form)) t))))
   (helpful-function #'test-foo-edebug))
 
@@ -245,13 +283,15 @@ symbol (not a form)."
     (should
      (eq
       (get-text-property 0 'face formatted)
-      'button)))
-  ;; Propertize mode maps.
-  (-let [formatted (helpful--format-docstring "`\\{python-mode-map}'")]
+      'button))))
+
+(ert-deftest helpful--format-docstring-mode-maps ()
+  "Ensure we propertize references to keymaps."
+  (-let [formatted (helpful--format-docstring "\\{python-mode-map}")]
     (should
      (s-contains-p "run-python" formatted)))
   ;; Handle non-existent mode maps gracefully.
-  (-let [formatted (helpful--format-docstring "`\\{no-such-mode-map}'")]
+  (-let [formatted (helpful--format-docstring "\\{no-such-mode-map}")]
     (should
      (s-contains-p "not currently defined" formatted))))
 
@@ -358,7 +398,8 @@ variables defined without `defvar'."
       (with-temp-buffer
         (insert "(defun test-foo-edebug-defn () 44)")
         (goto-char (point-min))
-        (shut-up
+	(cl-letf (((symbol-function #'message)
+		   (lambda (_format-string &rest _args))))
           (eval (eval-sexp-add-defvars (edebug-read-top-level-form)) t))
 
         (-let [(buf pos opened) (helpful--definition 'test-foo-edebug-defn t)]
@@ -581,6 +622,17 @@ associated a lambda with a keybinding."
     ;; Don't crash on anonymous functions in a keymap.
     (helpful--keymap-keys keymap)))
 
+(ert-deftest helpful--format-keymap--keyboard-macros ()
+  (let* ((keymap (make-keymap)))
+    ;; A keyboard macro can be a string or a vector.
+    (define-key keymap "a" "ABC")
+    (define-key keymap "b" [TAB])
+
+    (should
+     (equal
+      (helpful--format-keymap keymap)
+      "a  Keyboard Macro\nb  Keyboard Macro"))))
+
 (defun helpful--dummy-command ()
   (interactive))
 
@@ -881,6 +933,13 @@ find the source code."
   (should-not
    (helpful--convert-c-name 'Vgc_cons_percentage nil)))
 
+(ert-deftest helpful-symbol-unbound ()
+  "Ensure we inform the user if we're given an unbound symbol."
+  (should
+   (condition-case _
+       (helpful-symbol 'notboundtoanything)
+     ('user-error t))))
+
 (ert-deftest helpful--loads-autoload-symbol ()
   "When asked to describe an autoloaded symbol, just load it."
   ;; This test assumes that you haven't loaded tetris.el.gz in your
